@@ -15,6 +15,12 @@ defmodule Macro.ExternalTest do
   end
 end
 
+defmodule CustomIf do
+  def if(_cond, _expr) do
+    "custom if result"
+  end
+end
+
 defmodule MacroTest do
   use ExUnit.Case, async: true
   doctest Macro
@@ -208,7 +214,7 @@ defmodule MacroTest do
     end
 
     defp expand_once_and_clean(quoted, env) do
-      cleaner = &Keyword.drop(&1, [:counter])
+      cleaner = &Keyword.drop(&1, [:counter, :type_check])
 
       quoted
       |> Macro.expand_once(env)
@@ -270,7 +276,7 @@ defmodule MacroTest do
   end
 
   defp expand_and_clean(quoted, env) do
-    cleaner = &Keyword.drop(&1, [:counter])
+    cleaner = &Keyword.drop(&1, [:counter, :type_check])
 
     quoted
     |> Macro.expand(env)
@@ -373,14 +379,16 @@ defmodule MacroTest do
              """
     end
 
+    defp abc, do: [:a, :b, :c]
+
     test "with a pipeline on a single line" do
-      {result, formatted} = dbg_format([:a, :b, :c] |> tl() |> tl |> Kernel.hd())
+      {result, formatted} = dbg_format(abc() |> tl() |> tl |> Kernel.hd())
       assert result == :c
 
       assert formatted =~ "macro_test.exs"
 
       assert formatted =~ """
-             \n[:a, :b, :c] #=> [:a, :b, :c]
+             \nabc() #=> [:a, :b, :c]
              |> tl() #=> [:b, :c]
              |> tl #=> [:c]
              |> Kernel.hd() #=> :c
@@ -394,7 +402,7 @@ defmodule MacroTest do
     test "with a pipeline on multiple lines" do
       {result, formatted} =
         dbg_format(
-          [:a, :b, :c]
+          abc()
           |> tl()
           |> tl
           |> Kernel.hd()
@@ -405,7 +413,7 @@ defmodule MacroTest do
       assert formatted =~ "macro_test.exs"
 
       assert formatted =~ """
-             \n[:a, :b, :c] #=> [:a, :b, :c]
+             \nabc() #=> [:a, :b, :c]
              |> tl() #=> [:b, :c]
              |> tl #=> [:c]
              |> Kernel.hd() #=> :c
@@ -457,8 +465,32 @@ defmodule MacroTest do
              """
     end
 
+    test "with block of code" do
+      {result, formatted} =
+        dbg_format(
+          (
+            a = 1
+            b = a + 2
+            a + b
+          )
+        )
+
+      assert result == 4
+
+      assert formatted =~ "macro_test.exs"
+
+      assert formatted =~ """
+             Code block:
+             (
+               a = 1 #=> 1
+               b = a + 2 #=> 3
+               a + b #=> 4
+             )
+             """
+    end
+
     test "with case" do
-      list = [1, 2, 3]
+      list = List.flatten([1, 2, 3])
 
       {result, formatted} =
         dbg_format(
@@ -546,6 +578,238 @@ defmodule MacroTest do
              """
     end
 
+    test "if expression" do
+      x = true
+      map = %{a: 5, b: 1}
+
+      {result, formatted} =
+        dbg_format(
+          if true and x do
+            map[:a] * 2
+          else
+            map[:b]
+          end
+        )
+
+      assert result == 10
+
+      assert formatted =~ "macro_test.exs"
+
+      assert formatted =~ """
+             If condition:
+             true and x #=> true
+
+             If expression:
+             if true and x do
+               map[:a] * 2
+             else
+               map[:b]
+             end #=> 10
+             """
+    end
+
+    test "if expression without else" do
+      x = true
+      map = %{a: 5, b: 1}
+
+      {result, formatted} =
+        dbg_format(
+          if false and x do
+            map[:a] * 2
+          end
+        )
+
+      assert result == nil
+
+      assert formatted =~ "macro_test.exs"
+
+      assert formatted =~ """
+             If condition:
+             false and x #=> false
+
+             If expression:
+             if false and x do
+               map[:a] * 2
+             end #=> nil
+             """
+    end
+
+    test "custom if definition" do
+      import Kernel, except: [if: 2]
+      import CustomIf, only: [if: 2]
+
+      {result, formatted} =
+        dbg_format(
+          if true do
+            "something"
+          end
+        )
+
+      assert result == "custom if result"
+
+      assert formatted =~ """
+             if true do
+               "something"
+             end #=> "custom if result"
+             """
+    end
+
+    test "with with/1 (all clauses match)" do
+      opts = %{width: 10, height: 15}
+
+      {result, formatted} =
+        dbg_format(
+          with {:ok, width} <- Map.fetch(opts, :width),
+               double_width = width * 2,
+               IO.puts("just a side effect"),
+               {:ok, height} <- Map.fetch(opts, :height) do
+            {:ok, double_width * height}
+          end
+        )
+
+      assert result == {:ok, 300}
+
+      assert formatted =~ "macro_test.exs"
+
+      assert formatted =~ """
+             With clauses:
+             Map.fetch(opts, :width) #=> {:ok, 10}
+             width * 2 #=> 20
+             Map.fetch(opts, :height) #=> {:ok, 15}
+
+             With expression:
+             with {:ok, width} <- Map.fetch(opts, :width),
+                  double_width = width * 2,
+                  IO.puts("just a side effect"),
+                  {:ok, height} <- Map.fetch(opts, :height) do
+               {:ok, double_width * height}
+             end #=> {:ok, 300}
+             """
+    end
+
+    test "with with/1 (no else)" do
+      opts = %{width: 10}
+
+      {result, formatted} =
+        dbg_format(
+          with {:ok, width} <- Map.fetch(opts, :width),
+               {:ok, height} <- Map.fetch(opts, :height) do
+            {:ok, width * height}
+          end
+        )
+
+      assert result == :error
+
+      assert formatted =~ "macro_test.exs"
+
+      assert formatted =~ """
+             With clauses:
+             Map.fetch(opts, :width) #=> {:ok, 10}
+             Map.fetch(opts, :height) #=> :error
+
+             With expression:
+             with {:ok, width} <- Map.fetch(opts, :width),
+                  {:ok, height} <- Map.fetch(opts, :height) do
+               {:ok, width * height}
+             end #=> :error
+             """
+    end
+
+    test "with with/1 (else clause)" do
+      opts = %{width: 10}
+
+      {result, formatted} =
+        dbg_format(
+          with {:ok, width} <- Map.fetch(opts, :width),
+               {:ok, height} <- Map.fetch(opts, :height) do
+            width * height
+          else
+            :error -> 0
+          end
+        )
+
+      assert result == 0
+      assert formatted =~ "macro_test.exs"
+
+      assert formatted =~ """
+             With clauses:
+             Map.fetch(opts, :width) #=> {:ok, 10}
+             Map.fetch(opts, :height) #=> :error
+
+             With expression:
+             with {:ok, width} <- Map.fetch(opts, :width),
+                  {:ok, height} <- Map.fetch(opts, :height) do
+               width * height
+             else
+               :error -> 0
+             end #=> 0
+             """
+    end
+
+    test "with with/1 (guard)" do
+      opts = %{width: 10, height: 0.0}
+
+      {result, formatted} =
+        dbg_format(
+          with {:ok, width} when is_integer(width) <- Map.fetch(opts, :width),
+               {:ok, height} when is_integer(height) <- Map.fetch(opts, :height) do
+            width * height
+          else
+            _ -> nil
+          end
+        )
+
+      assert result == nil
+      assert formatted =~ "macro_test.exs"
+
+      assert formatted =~ """
+             With clauses:
+             Map.fetch(opts, :width) #=> {:ok, 10}
+             Map.fetch(opts, :height) #=> {:ok, 0.0}
+
+             With expression:
+             with {:ok, width} when is_integer(width) <- Map.fetch(opts, :width),
+                  {:ok, height} when is_integer(height) <- Map.fetch(opts, :height) do
+               width * height
+             else
+               _ -> nil
+             end #=> nil
+             """
+    end
+
+    test "with with/1 (guard in else)" do
+      opts = %{}
+
+      {result, _formatted} =
+        dbg_format(
+          with {:ok, width} <- Map.fetch(opts, :width) do
+            width
+          else
+            other when is_integer(other) -> :int
+            other when is_atom(other) -> :atom
+          end
+        )
+
+      assert result == :atom
+    end
+
+    test "with with/1 respects the WithClauseError" do
+      value = Enum.random([:unexpected])
+
+      error =
+        assert_raise WithClauseError, fn ->
+          dbg(
+            with :ok <- value do
+              true
+            else
+              :error -> false
+            end
+          )
+        end
+
+      assert error.term == :unexpected
+    end
+
     test "with \"syntax_colors: []\" it doesn't print any color sequences" do
       {_result, formatted} = dbg_format("hello")
       refute formatted =~ "\e["
@@ -574,40 +838,8 @@ defmodule MacroTest do
       assert Macro.to_string(quote do: hello(world)) == "hello(world)"
     end
 
-    test "large number literals" do
-      # with quote
-      assert Macro.to_string(quote do: 576_460_752_303_423_455) == "576_460_752_303_423_455"
-      assert Macro.to_string(quote do: -576_460_752_303_423_455) == "-576_460_752_303_423_455"
-
-      # without quote
-      assert Macro.to_string(576_460_752_303_423_455) == "576_460_752_303_423_455"
-      assert Macro.to_string(-576_460_752_303_423_455) == "-576_460_752_303_423_455"
-    end
-
-    defmodule HTML do
-      defstruct [:string]
-
-      defimpl Inspect do
-        def inspect(%{string: string}, _) do
-          "~HTML[#{string}]"
-        end
-      end
-    end
-
-    defmacro sigil_HTML({:<<>>, _, [string]}, []) do
-      Macro.escape(%HTML{string: string})
-    end
-
-    test "sigils" do
-      assert Macro.to_string(quote(do: ~HTML[hi])) == ~S/~HTML[hi]/
-
-      assert Macro.to_string(
-               quote do
-                 ~HTML"""
-                 hi
-                 """
-               end
-             ) == ~s[~HTML"""\nhi\n"""]
+    test "converts invalid AST with inspect" do
+      assert Macro.to_string(1..3) == "1..3"
     end
   end
 
@@ -940,12 +1172,16 @@ defmodule MacroTest do
     end
 
     test "range" do
-      assert macro_to_string(quote(do: unquote(-1..+2))) == "-1..2"
+      assert macro_to_string(quote(do: -1..+2)) == "-1..+2"
       assert macro_to_string(quote(do: Foo.integer()..3)) == "Foo.integer()..3"
-      assert macro_to_string(quote(do: unquote(-1..+2//-3))) == "-1..2//-3"
+      assert macro_to_string(quote(do: -1..+2//-3)) == "-1..+2//-3"
 
       assert macro_to_string(quote(do: Foo.integer()..3//Bar.bat())) ==
                "Foo.integer()..3//Bar.bat()"
+
+      # invalid AST
+      assert macro_to_string(-1..+2) == "-1..2"
+      assert macro_to_string(-1..+2//-3) == "-1..2//-3"
     end
 
     test "when" do
@@ -1311,22 +1547,34 @@ defmodule MacroTest do
     Macro.postwalk(ast, [], &{&1, [&1 | &2]}) |> elem(1) |> Enum.reverse()
   end
 
-  test "struct!/2 expands structs multiple levels deep" do
+  test "struct_info!/2 expands structs multiple levels deep" do
     defmodule StructBang do
       defstruct [:a, :b]
 
-      assert Macro.struct!(StructBang, __ENV__) == %{__struct__: StructBang, a: nil, b: nil}
+      assert Macro.struct_info!(StructBang, __ENV__) == [
+               %{field: :a, default: nil},
+               %{field: :b, default: nil}
+             ]
 
       def within_function do
-        assert Macro.struct!(StructBang, __ENV__) == %{__struct__: StructBang, a: nil, b: nil}
+        assert Macro.struct_info!(StructBang, __ENV__) == [
+                 %{field: :a, default: nil},
+                 %{field: :b, default: nil}
+               ]
       end
 
       defmodule Nested do
-        assert Macro.struct!(StructBang, __ENV__) == %{__struct__: StructBang, a: nil, b: nil}
+        assert Macro.struct_info!(StructBang, __ENV__) == [
+                 %{field: :a, default: nil},
+                 %{field: :b, default: nil}
+               ]
       end
     end
 
-    assert Macro.struct!(StructBang, __ENV__) == %{__struct__: StructBang, a: nil, b: nil}
+    assert Macro.struct_info!(StructBang, __ENV__) == [
+             %{field: :a, default: nil},
+             %{field: :b, default: nil}
+           ]
   end
 
   test "prewalker/1" do
