@@ -14,8 +14,8 @@ defmodule TypespecTest do
           unquote(block)
         end
 
-      :code.delete(TypespecSample)
       :code.purge(TypespecSample)
+      :code.delete(TypespecSample)
       bytecode
     end
   end
@@ -75,6 +75,14 @@ defmodule TypespecTest do
                        @type my_type :: %URI.t(){}
                      end
                    end
+
+      assert_raise Kernel.TypespecError,
+                   ~r"unexpected expression in typespec: t\.Foo",
+                   fn ->
+                     test_module do
+                       @type my_type :: t.Foo
+                     end
+                   end
     end
 
     test "invalid function specification" do
@@ -120,7 +128,7 @@ defmodule TypespecTest do
 
     test "redefined type" do
       assert_raise Kernel.TypespecError,
-                   ~r"type foo/0 is already defined in .*test/elixir/typespec_test.exs:126",
+                   ~r"type foo/0 is already defined in .*test/elixir/typespec_test.exs:134",
                    fn ->
                      test_module do
                        @type foo :: atom
@@ -129,7 +137,7 @@ defmodule TypespecTest do
                    end
 
       assert_raise Kernel.TypespecError,
-                   ~r"type foo/2 is already defined in .*test/elixir/typespec_test.exs:136",
+                   ~r"type foo/2 is already defined in .*test/elixir/typespec_test.exs:144",
                    fn ->
                      test_module do
                        @type foo :: atom
@@ -139,7 +147,7 @@ defmodule TypespecTest do
                    end
 
       assert_raise Kernel.TypespecError,
-                   ~r"type foo/0 is already defined in .*test/elixir/typespec_test.exs:145",
+                   ~r"type foo/0 is already defined in .*test/elixir/typespec_test.exs:153",
                    fn ->
                      test_module do
                        @type foo :: atom
@@ -406,7 +414,7 @@ defmodule TypespecTest do
 
       assert_raise Kernel.TypespecError, ~r"invalid binary specification", fn ->
         test_module do
-          @type my_type :: <<_::atom>>
+          @type my_type :: <<_::atom()>>
         end
       end
 
@@ -547,7 +555,7 @@ defmodule TypespecTest do
     end
 
     test "@type with undefined struct" do
-      assert_raise ArgumentError, ~r"ThisModuleDoesNotExist.__struct__/0 is undefined", fn ->
+      assert_raise ArgumentError, ~r"ThisModuleDoesNotExist.__struct__/1 is undefined", fn ->
         test_module do
           @type my_type :: %ThisModuleDoesNotExist{}
         end
@@ -710,7 +718,8 @@ defmodule TypespecTest do
           @type my_type :: (... -> any)
         end
 
-      assert [type: {:my_type, {:type, _, :fun, []}, []}] = types(bytecode)
+      assert [type: {:my_type, {:type, _, :fun, [{:type, _, :any}, {:type, _, :any, []}]}, []}] =
+               types(bytecode)
     end
 
     test "@type with a fun with multiple arguments and return type" do
@@ -818,7 +827,7 @@ defmodule TypespecTest do
 
       bytecode =
         test_module do
-          Module.eval_quoted(__MODULE__, quoted)
+          Code.eval_quoted(quoted, [], module: __MODULE__)
         end
 
       assert [type: {:my_type, {:atom, _, :foo}, []}] = types(bytecode)
@@ -840,6 +849,19 @@ defmodule TypespecTest do
       assert [{:atom, _, Keyword}, {:atom, _, :t}, [{:var, _, :value}]] = kw_with_value_args
     end
 
+    test "@type with macro in alias" do
+      bytecode =
+        test_module do
+          defmacro module() do
+            quote do: __MODULE__
+          end
+
+          @type my_type :: module().Foo
+        end
+
+      assert [type: {:my_type, {:atom, _, TypespecTest.TypespecSample.Foo}, []}] = types(bytecode)
+    end
+
     test "@type with a reserved signature" do
       assert_raise Kernel.TypespecError,
                    ~r"type required\/1 is a reserved type and it cannot be defined",
@@ -1070,8 +1092,8 @@ defmodule TypespecTest do
                 {TypeModuleAttributes, _}}
              ] = TypeModuleAttributes.typep2()
     after
-      :code.delete(TypeModuleAttributes)
       :code.purge(TypeModuleAttributes)
+      :code.delete(TypeModuleAttributes)
     end
 
     test "@spec, @callback, and @macrocallback as module attributes" do
@@ -1129,8 +1151,8 @@ defmodule TypespecTest do
                 {SpecModuleAttributes, _}}
              ] = SpecModuleAttributes.macrocallback()
     after
-      :code.delete(SpecModuleAttributes)
       :code.purge(SpecModuleAttributes)
+      :code.delete(SpecModuleAttributes)
     end
 
     test "@callback" do
@@ -1223,7 +1245,7 @@ defmodule TypespecTest do
 
       bytecode =
         test_module do
-          Module.eval_quoted(__MODULE__, quoted)
+          Code.eval_quoted(quoted, [], module: __MODULE__)
         end
 
       types = types(bytecode)
@@ -1265,7 +1287,7 @@ defmodule TypespecTest do
 
           def foo(), do: 1
           def foo(arg), do: arg
-          Module.eval_quoted(__MODULE__, quote(do: (unquote_splicing(quoted))))
+          Code.eval_quoted(quote(do: (unquote_splicing(quoted))), [], module: __MODULE__)
         end
 
       specs =
@@ -1283,32 +1305,32 @@ defmodule TypespecTest do
     end
 
     test "spec_to_quoted with maps with __struct__ key" do
-      defmodule A do
+      defmodule StructA do
         defstruct [:key]
       end
 
-      defmodule B do
+      defmodule StructB do
         defstruct [:key]
       end
 
       bytecode =
         test_module do
-          @spec single_struct(%A{}) :: :ok
+          @spec single_struct(%StructA{}) :: :ok
           def single_struct(arg), do: {:ok, arg}
 
-          @spec single_struct_key(%{__struct__: A}) :: :ok
+          @spec single_struct_key(%{__struct__: StructA}) :: :ok
           def single_struct_key(arg), do: {:ok, arg}
 
           @spec single_struct_key_type(%{__struct__: atom()}) :: :ok
           def single_struct_key_type(arg), do: {:ok, arg}
 
-          @spec union_struct(%A{} | %B{}) :: :ok
+          @spec union_struct(%StructA{} | %StructB{}) :: :ok
           def union_struct(arg), do: {:ok, arg}
 
-          @spec union_struct_key(%{__struct__: A | B}) :: :ok
+          @spec union_struct_key(%{__struct__: StructA | StructB}) :: :ok
           def union_struct_key(arg), do: {:ok, arg}
 
-          @spec union_struct_key_type(%{__struct__: atom() | A | binary()}) :: :ok
+          @spec union_struct_key_type(%{__struct__: atom() | StructA | binary()}) :: :ok
           def union_struct_key_type(arg), do: {:ok, arg}
         end
 
@@ -1323,26 +1345,26 @@ defmodule TypespecTest do
 
       assert Code.Typespec.spec_to_quoted(:single_struct, ast_single_struct)
              |> Macro.to_string() ==
-               "single_struct(%TypespecTest.A{key: term()}) :: :ok"
+               "single_struct(%TypespecTest.StructA{key: term()}) :: :ok"
 
       assert Code.Typespec.spec_to_quoted(:single_struct_key, ast_single_struct_key)
              |> Macro.to_string() ==
-               "single_struct_key(%TypespecTest.A{}) :: :ok"
+               "single_struct_key(%TypespecTest.StructA{}) :: :ok"
 
       assert Code.Typespec.spec_to_quoted(:single_struct_key_type, ast_single_struct_key_type)
              |> Macro.to_string() ==
                "single_struct_key_type(%{__struct__: atom()}) :: :ok"
 
       assert Code.Typespec.spec_to_quoted(:union_struct, ast_union_struct) |> Macro.to_string() ==
-               "union_struct(%TypespecTest.A{key: term()} | %TypespecTest.B{key: term()}) :: :ok"
+               "union_struct(%TypespecTest.StructA{key: term()} | %TypespecTest.StructB{key: term()}) :: :ok"
 
       assert Code.Typespec.spec_to_quoted(:union_struct_key, ast_union_struct_key)
              |> Macro.to_string() ==
-               "union_struct_key(%{__struct__: TypespecTest.A | TypespecTest.B}) :: :ok"
+               "union_struct_key(%{__struct__: TypespecTest.StructA | TypespecTest.StructB}) :: :ok"
 
       assert Code.Typespec.spec_to_quoted(:union_struct_key_type, ast_union_struct_key_type)
              |> Macro.to_string() ==
-               "union_struct_key_type(%{__struct__: atom() | TypespecTest.A | binary()}) :: :ok"
+               "union_struct_key_type(%{__struct__: atom() | TypespecTest.StructA | binary()}) :: :ok"
     end
 
     test "non-variables are given as arguments" do
@@ -1486,7 +1508,7 @@ defmodule TypespecTest do
 
       bytecode =
         test_module do
-          Module.eval_quoted(__MODULE__, quoted)
+          Code.eval_quoted(quoted, [], module: __MODULE__)
         end
 
       types = types(bytecode)
@@ -1520,9 +1542,6 @@ defmodule TypespecTest do
             assert ast_string ==
                      "@type literal_struct_all_fields_key_type() :: %TypespecTest.SomeStruct{key: integer()}"
 
-          {:built_in_fun, _, _} ->
-            assert ast_string == "@type built_in_fun() :: (... -> any())"
-
           {:built_in_nonempty_list, _, _} ->
             assert ast_string == "@type built_in_nonempty_list() :: [...]"
 
