@@ -10,6 +10,10 @@ defmodule ExUnit.DiffTest do
     defstruct [:name, :age]
   end
 
+  defmodule Customer do
+    defstruct [:address, :age, :first_name, :language, :last_name, :notifications]
+  end
+
   defmodule Person do
     defstruct [:age]
   end
@@ -126,10 +130,11 @@ defmodule ExUnit.DiffTest do
     assert env_binding == expected_binding
   end
 
+  @terminal_width 80
   defp to_diff(side, sign) do
     side
     |> Diff.to_algebra(&diff_wrapper(&1, sign))
-    |> Algebra.format(:infinity)
+    |> Algebra.format(@terminal_width)
     |> IO.iodata_to_binary()
   end
 
@@ -575,6 +580,12 @@ defmodule ExUnit.DiffTest do
     refute_diff(%{a: 1, b: 12} == %{a: 1, b: 2}, "%{a: 1, b: -1-2}", "%{a: 1, b: 2}")
     refute_diff(%{a: 1, b: 2} == %{a: 1, b: 12}, "%{a: 1, b: 2}", "%{a: 1, b: +1+2}")
     refute_diff(%{a: 1, b: 2} == %{a: 1, c: 2}, "%{a: 1, -b: 2-}", "%{a: 1, +c: 2+}")
+
+    refute_diff(
+      %{name: {:a, :b, :c}} == :error,
+      "-%{name: {:a, :b, :c}}",
+      "+:error+"
+    )
   end
 
   test "structs" do
@@ -682,13 +693,49 @@ defmodule ExUnit.DiffTest do
     refute_diff(
       %User{age: %Date{}} = struct,
       ~s/%ExUnit.DiffTest.User{age: %-Date-{}}/,
-      ~s/%ExUnit.DiffTest.User{age: %+DateTime+{calendar: Calendar.ISO, day: 30, hour: 13, microsecond: {253158, 6}, minute: 49, month: 7, second: 59, std_offset: 0, time_zone: "Etc\/UTC", utc_offset: 0, year: 2020, zone_abbr: "UTC"}, name: nil}/
+      """
+      %ExUnit.DiffTest.User{
+        age: %+DateTime+{
+          calendar: Calendar.ISO,
+          day: 30,
+          hour: 13,
+          microsecond: {253158, 6},
+          minute: 49,
+          month: 7,
+          second: 59,
+          std_offset: 0,
+          time_zone: "Etc\/UTC",
+          utc_offset: 0,
+          year: 2020,
+          zone_abbr: "UTC"
+        },
+        name: nil
+      }\
+      """
     )
 
     refute_diff(
       %{age: %Date{}} = struct,
       ~s/%{age: %-Date-{}}/,
-      ~s/%ExUnit.DiffTest.User{age: %+DateTime+{calendar: Calendar.ISO, day: 30, hour: 13, microsecond: {253158, 6}, minute: 49, month: 7, second: 59, std_offset: 0, time_zone: "Etc\/UTC", utc_offset: 0, year: 2020, zone_abbr: "UTC"}, name: nil}/
+      """
+      %ExUnit.DiffTest.User{
+        age: %+DateTime+{
+          calendar: Calendar.ISO,
+          day: 30,
+          hour: 13,
+          microsecond: {253158, 6},
+          minute: 49,
+          month: 7,
+          second: 59,
+          std_offset: 0,
+          time_zone: \"Etc/UTC\",
+          utc_offset: 0,
+          year: 2020,
+          zone_abbr: \"UTC\"
+        },
+        name: nil
+      }\
+      """
     )
   end
 
@@ -836,6 +883,106 @@ defmodule ExUnit.DiffTest do
     )
   end
 
+  test "maps in lists" do
+    map = %{
+      "address" => %{
+        "street" => "123 Main St",
+        "zip" => "62701"
+      },
+      "age" => 30,
+      "first_name" => "John",
+      "language" => "en-US",
+      "last_name" => "Doe",
+      "notifications" => true
+    }
+
+    refute_diff(
+      [map] == [],
+      """
+      [
+        -%{
+          "address" => %{"street" => "123 Main St", "zip" => "62701"},
+          "age" => 30,
+          "first_name" => "John",
+          "language" => "en-US",
+          "last_name" => "Doe",
+          "notifications" => true
+        }-
+      ]\
+      """,
+      "[]"
+    )
+
+    refute_diff(
+      [] == [map],
+      "[]",
+      """
+      [
+        +%{
+          "address" => %{"street" => "123 Main St", "zip" => "62701"},
+          "age" => 30,
+          "first_name" => "John",
+          "language" => "en-US",
+          "last_name" => "Doe",
+          "notifications" => true
+        }+
+      ]\
+      """
+    )
+
+    assert_diff([map] == [map], [])
+  end
+
+  test "structs in lists" do
+    customer = %Customer{
+      address: %{
+        "street" => "123 Main St",
+        "zip" => "62701"
+      },
+      age: 30,
+      first_name: "John",
+      language: "en-US",
+      last_name: "Doe",
+      notifications: true
+    }
+
+    refute_diff(
+      [customer] == [],
+      """
+      [
+        -%ExUnit.DiffTest.Customer{
+          address: %{"street" => "123 Main St", "zip" => "62701"},
+          age: 30,
+          first_name: "John",
+          language: "en-US",
+          last_name: "Doe",
+          notifications: true
+        }-
+      ]\
+      """,
+      "[]"
+    )
+
+    refute_diff(
+      [] == [customer],
+      "[]",
+      """
+      [
+        +%ExUnit.DiffTest.Customer{
+          address: %{"street" => "123 Main St", "zip" => "62701"},
+          age: 30,
+          first_name: "John",
+          language: "en-US",
+          last_name: "Doe",
+          notifications: true
+        }+
+      ]\
+      """
+    )
+
+    assert_diff([customer] == [customer], [])
+  end
+
   test "maps and structs with escaped values" do
     refute_diff(
       %User{age: {1, 2, 3}} = %User{age: {1, 2, 4}},
@@ -986,6 +1133,12 @@ defmodule ExUnit.DiffTest do
       "-<<trap::binary-size(3)>> <> \"baz\"-",
       "+\"foobar\"+"
     )
+
+    refute_diff(
+      "hello " <> <<_::binary-size(6)>> = "hello world",
+      "\"hello \" <> -<<_::binary-size(6)>>-",
+      "\"hello +world+\""
+    )
   end
 
   test "underscore" do
@@ -1082,8 +1235,18 @@ defmodule ExUnit.DiffTest do
 
     refute_diff(
       {ref1, ref2} == {ref2, ref1},
-      "{-#{inspect_ref1}-, -#{inspect_ref2}-}",
-      "{+#{inspect_ref2}+, +#{inspect_ref1}+}"
+      """
+      {
+        -#{inspect_ref1}-,
+        -#{inspect_ref2}-
+      }\
+      """,
+      """
+      {
+        +#{inspect_ref2}+,
+        +#{inspect_ref1}+
+      }\
+      """
     )
 
     refute_diff(
@@ -1100,7 +1263,16 @@ defmodule ExUnit.DiffTest do
 
     refute_diff(ref1 == :a, "-#{inspect_ref1}-", "+:a+")
     refute_diff({ref1, ref2} == :a, "-{#{inspect_ref1}, #{inspect_ref2}}", "+:a+")
-    refute_diff(%{ref1 => ref2} == :a, "-%{#{inspect_ref1} => #{inspect_ref2}}", "+:a+")
+
+    refute_diff(
+      %{ref1 => ref2} == :a,
+      """
+      -%{
+        #{inspect_ref1} => #{inspect_ref2}
+      }\
+      """,
+      "+:a+"
+    )
 
     refute_diff(
       %Opaque{data: ref1} == :a,
@@ -1141,7 +1313,16 @@ defmodule ExUnit.DiffTest do
     refute_diff(identity == :a, "-#{inspect}-", "+:a+")
     refute_diff({identity, identity} == :a, "-{#{inspect}, #{inspect}}", "+:a+")
     refute_diff({identity, :a} == {:a, identity}, "{-#{inspect}-, -:a-}", "{+:a+, +#{inspect}+}")
-    refute_diff(%{identity => identity} == :a, "-%{#{inspect} => #{inspect}}", "+:a+")
+
+    refute_diff(
+      %{identity => identity} == :a,
+      """
+      -%{
+        #{inspect} => #{inspect}
+      }-\
+      """,
+      "+:a+"
+    )
 
     refute_diff(
       (&String.to_charlist/1) == (&String.unknown/1),
