@@ -363,7 +363,7 @@ defmodule IEx.Autocomplete do
       end)
 
     entries =
-      for {key, _value} <- pairs,
+      for key when key != :__struct__ <- Map.keys(pairs),
           name = Atom.to_string(key),
           if(hint == "",
             do: not String.starts_with?(name, "_"),
@@ -685,7 +685,7 @@ defmodule IEx.Autocomplete do
 
   defp hidden_fun?({name, arity}, docs) do
     case Enum.find(docs, &match?({{_, ^name, ^arity}, _, _, _, _}, &1)) do
-      nil -> hd(Atom.to_charlist(name)) == ?_
+      nil -> match?([?_ | _], Atom.to_charlist(name))
       {_, _, _, :hidden, _} -> true
       {_, _, _, _, _} -> false
     end
